/**
 * Copyright (C) 2009-2010  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.definition.activity.ConnectorExecutor;
import org.ow2.bonita.deployment.DeploymentRuntimeException;
import org.ow2.bonita.facade.def.InternalProcessDefinition;
import org.ow2.bonita.facade.def.element.AttachmentDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.element.IncomingEventDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.EventProcessDefinition;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.impl.FacadeUtil;
import org.ow2.bonita.facade.runtime.InitialAttachment;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.impl.AttachmentInstanceImpl;
import org.ow2.bonita.facade.runtime.impl.InternalProcessInstance;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.event.EventConstants;
import org.ow2.bonita.runtime.event.IncomingEventInstance;
import org.ow2.bonita.runtime.event.OutgoingEventInstance;
import org.ow2.bonita.runtime.model.Execution;
import org.ow2.bonita.services.EventService;
import org.ow2.bonita.services.LargeDataRepository;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.type.Variable;

/**
 * 
 * @author Charles Souillard, Matthieu Chaffotte
 *
 */
public class ProcessUtil {

  static final Logger LOG = Logger.getLogger(ProcessUtil.class.getName());

  public static void removeAllInstanceEvents(final ProcessInstance instance) {
    final EventService eventService = EnvTool.getEventService();
    eventService.removeSubscriptions(instance.getUUID());
    final Set<OutgoingEventInstance> outgoings = eventService.getOutgoingEvents(instance.getUUID());
    for (OutgoingEventInstance outgoing : outgoings) {
      eventService.removeEvent(outgoing);
    }
  }

  public static void removeSubProcessEvents(final ProcessInstance instance) {
    final Querier database = EnvTool.getJournalQueriers();
    final InternalProcessDefinition process = database.getProcess(instance.getProcessDefinitionUUID());
    if (process != null) {
      final EventService eventService = EnvTool.getEventService();
      for (EventProcessDefinition eventSubProcess : process.getEventSubProcesses()) {
        final InternalProcessDefinition eventProcess = EnvTool.getJournalQueriers().getProcess(eventSubProcess.getName(), eventSubProcess.getVersion());
        ActivityDefinition startEvent = getStartEvent(eventProcess);
        if (startEvent.isTimer()) {
          final InternalProcessDefinition def = database.getProcess(startEvent.getProcessDefinitionUUID());
          Set<OutgoingEventInstance> delete = eventService.getOutgoingEvents(BonitaConstants.TIMER_EVENT_PREFIX + startEvent.getUUID(), def.getName(), startEvent.getName(), null);
          if (!delete.isEmpty()) {
            eventService.removeEvent(delete.iterator().next());
          }
        }
      }
    }
  }

  public static void removeInternalInstanceEvents(final ProcessInstanceUUID instanceUUID) {
    final EventService eventService = EnvTool.getEventService();
    //delete timers, async events, deadlines
    eventService.removeSubscriptions(instanceUUID);
    Set<OutgoingEventInstance> outgoings = eventService.getOutgoingEvents(instanceUUID);
    for (OutgoingEventInstance outgoing : outgoings) {
      final String name = outgoing.getName();
      if (name.startsWith(BonitaConstants.TIMER_EVENT_PREFIX)
          || name.startsWith(BonitaConstants.DEADLINE_EVENT_PREFIX)
          || name.startsWith(BonitaConstants.ASYNC_EVENT_PREFIX)) {
        eventService.removeEvent(outgoing);
      }
    }
  }

  public static Date getTimerDate(final String condition, final ProcessDefinitionUUID definitionUUID, final long lastTime) throws GroovyException {
    return getTimerDate(condition, definitionUUID, null, lastTime);
  }

  public static Date getTimerDate(final String condition, final ActivityInstanceUUID activityUUID) throws GroovyException {
    return getTimerDate(condition, null, activityUUID, 0);
  }

  private static Date getTimerDate(final String condition, final ProcessDefinitionUUID definitionUUID, final ActivityInstanceUUID activityUUID, final long lastTime) throws GroovyException {
    Misc.checkArgsNotNull(condition);
    Date dueDate = null;
    String evaluatedCondition = condition;
    if (Misc.isJustAGroovyExpression(condition)) {
      Object value = null;
      if (activityUUID != null) {
        value = GroovyUtil.evaluate(condition, null, activityUUID, false, false);
      } else {
        Map<String, Object> context = new HashMap<String, Object>();
        context.put(BonitaConstants.TIMER_LAST_EXECUTION, lastTime);
        value = GroovyUtil.evaluate(condition, context, definitionUUID, false);
      }
      if (value instanceof Date) {
        dueDate = (Date) value;
      } else {
        if (value instanceof String) {
          evaluatedCondition = (String) value;
        } else if (value instanceof Long) {
          evaluatedCondition = ((Long)value).toString();
        } else if (value instanceof Integer) {
          evaluatedCondition = ((Integer)value).toString();
        }
      }
    }
    if (dueDate == null) {
      try {
        Long date = Long.parseLong(evaluatedCondition);
        if (activityUUID != null) {
          dueDate = new Date(System.currentTimeMillis() + date);
        } else if (lastTime < 0 ) {
          dueDate = new Date(System.currentTimeMillis() + date);
        } else {
          dueDate = new Date(date);
        }
      } catch (final NumberFormatException e1) {
        try {
          if (dueDate == null) {
            dueDate = DateUtil.parseDate(evaluatedCondition);
          }
        } catch (final IllegalArgumentException e2) {
          throw new BonitaRuntimeException("Timer condition '" + evaluatedCondition + "' is neither a Long nor a formatted date", e2);
        }
      }
    }
    return dueDate;
  }

  public static Execution createProcessInstance(final ProcessDefinitionUUID processUUID, Map<String, Object> variables, Collection<InitialAttachment> attachments, final ProcessInstanceUUID parentInstanceUUID,
      final ProcessInstanceUUID rootInstanceUUID, final ActivityDefinitionUUID activityUUID, final ActivityInstanceUUID parentActivityUUID) throws ProcessNotFoundException {
    final InternalProcessDefinition process = FacadeUtil.getProcessDefinition(processUUID);

    Execution rootExecution = createProcessInstance(process, rootInstanceUUID, activityUUID);
    InternalProcessInstance instance = rootExecution.getInstance();

    final Recorder recorder = EnvTool.getRecorder();
    Map<String, Variable> givenVariables = VariableUtil.createVariableMap(processUUID, variables);
    Map<String, Variable> processVariables = VariableUtil.createVariables(process.getDataFields(), null);
    Map<String, Variable> initialVariables = new HashMap<String, Variable>();
    if (processVariables != null) {
      initialVariables.putAll(processVariables);
    }
    if (givenVariables != null) {
      initialVariables.putAll(givenVariables);
    }

    instance.setParentUUIDs(parentInstanceUUID, parentActivityUUID);
    instance.setInitialVaribales(initialVariables);
    String initiator = EnvTool.getUserId();
    if (parentInstanceUUID != null) {
      InternalProcessInstance parentInstance = EnvTool.getJournalQueriers().getProcessInstance(parentInstanceUUID);
      initiator = parentInstance.getStartedBy();
    }
    recorder.recordInstanceStarted(instance, initiator);

    final Set<String> runtimeAttachmentNames = new HashSet<String>();
    if (attachments != null) {
      for (InitialAttachment attachment : attachments) {
        runtimeAttachmentNames.add(attachment.getName());
        AttachmentInstanceImpl newAttachment = new AttachmentInstanceImpl(attachment, instance.getUUID(), EnvTool.getUserId(), new Date());
        ProcessUtil.addAttachment(instance, newAttachment, attachment.getContent());
      }
    }
    LargeDataRepository ldr = EnvTool.getLargeDataRepository();
    for (AttachmentDefinition attachmentDefinition : process.getAttachments().values()) {
      if (!runtimeAttachmentNames.contains(attachmentDefinition.getName())) {
        //do not override parameter attachments
        byte[] initialValue = null;
        if (attachmentDefinition.getFilePath() != null) {
          initialValue = ldr.getData(byte[].class, Misc.getAttachmentCategories(processUUID), attachmentDefinition.getFilePath());
        }
        AttachmentInstanceImpl attachmentInstance = new AttachmentInstanceImpl(attachmentDefinition.getName(), instance.getUUID(), EnvTool.getUserId(), new Date());
        attachmentInstance.setLabel(attachmentDefinition.getLabel());
        attachmentInstance.setDescription(attachmentDefinition.getDescription());
        attachmentInstance.setFileName(attachmentDefinition.getFileName());
        ProcessUtil.addAttachment(instance, attachmentInstance, initialValue);
      }
    }
    ConnectorExecutor.executeConnectors(rootExecution, HookDefinition.Event.instanceOnStart);

    return rootExecution;
  }

  public static void startEventSubProcesses(final ProcessInstance instance)
  throws ProcessNotFoundException {
    final ProcessInstanceUUID instanceUUID = instance.getUUID();
    final InternalProcessDefinition process = FacadeUtil.getProcessDefinition(instance.getProcessDefinitionUUID());
    for (EventProcessDefinition eventSubProcess : process.getEventSubProcesses()) {
      final InternalProcessDefinition eventProcess = EnvTool.getJournalQueriers().getProcess(eventSubProcess.getName(), eventSubProcess.getVersion());
      if (eventProcess == null) {
        throw new ProcessNotFoundException("bs_PU_1", eventSubProcess.getName(), eventSubProcess.getVersion());
      }
      ActivityDefinition startEvent = getStartEvent(eventProcess);
      if (startEvent.isTimer()) {
        final EventService eventService = EnvTool.getEventService();
        final String eventName = BonitaConstants.TIMER_EVENT_PREFIX + startEvent.getUUID();
        final String condition = startEvent.getTimerCondition();
        try {
          final OutgoingEventInstance outgoing = new OutgoingEventInstance(eventName, eventProcess.getName(), startEvent.getName(), null, null, null, -1);
          eventService.fire(outgoing);
          final Date date = (Date) getTimerDate(condition, process.getUUID(), -1);
          long enableTime = date.getTime();
          IncomingEventInstance timerEventInstance = new IncomingEventInstance(eventName, condition, null, startEvent.getUUID(), null, eventProcess.getName(), startEvent.getName(), null, EventConstants.TIMER_START_EVENT, enableTime, false);
          timerEventInstance.setPermanent(true);
          timerEventInstance.setEventSubProcessRootInstanceUUID(instanceUUID);
          eventService.subscribe(timerEventInstance);
        } catch (GroovyException e) {
          throw new DeploymentRuntimeException(e.getMessage(), e.getCause());
        }
      } else if (startEvent.isCatchingSignalEvent()) {
        final String eventName = startEvent.getTimerCondition();
        IncomingEventInstance signalEventInstance = new IncomingEventInstance(eventName, null, null, startEvent.getUUID(), null, eventProcess.getName(), startEvent.getName(), null, EventConstants.SIGNAL_START_EVENT, -1, false);
        signalEventInstance.setPermanent(true);
        signalEventInstance.setEventSubProcessRootInstanceUUID(instanceUUID);
        final EventService eventService = EnvTool.getEventService();
        eventService.subscribe(signalEventInstance);
      } else if (startEvent.isReceiveEvent()) {
        final IncomingEventDefinition event = startEvent.getIncomingEvent();
        if (event != null) {
          final EventService eventService = EnvTool.getEventService();
          IncomingEventInstance incomingEventInstance = new IncomingEventInstance(event.getName(), event.getExpression(), null, startEvent.getUUID(), null, eventProcess.getName(), startEvent.getName(), null, EventConstants.MESSAGE_START_EVENT, System.currentTimeMillis(), false);
          incomingEventInstance.setPermanent(true);
          incomingEventInstance.setEventSubProcessRootInstanceUUID(instanceUUID);
          eventService.subscribe(incomingEventInstance);
        }
      } else if (startEvent.isCatchingErrorEvent()) {
        final String errorCode = startEvent.getTimerCondition();
        StringBuilder builder = new StringBuilder(startEvent.getName());
        builder.append(EventConstants.SEPARATOR);
        if (errorCode != null) {
          builder.append(errorCode);
        } else {
          builder.append("all");
        }
        IncomingEventInstance errorEventInstance = new IncomingEventInstance(builder.toString(), null, null, startEvent.getUUID(), null, eventProcess.getName(), startEvent.getName(), null, EventConstants.ERROR_START_EVENT, -1, true);
        errorEventInstance.setPermanent(true);
        errorEventInstance.setEventSubProcessRootInstanceUUID(instanceUUID);
        final EventService eventService = EnvTool.getEventService();
        eventService.subscribe(errorEventInstance);
      }
    }
  }

  private static ActivityDefinition getStartEvent(InternalProcessDefinition eventProcess) {
    for (ActivityDefinition activity : eventProcess.getActivities()) {
      if (activity.getIncomingTransitions().isEmpty()) {
        return activity;
      }
    }
    return null;
  }

  public static void addAttachment(InternalProcessInstance processInstance, AttachmentInstanceImpl attachmentInstance, byte[] value) {
    LargeDataRepository largeDataRepository = EnvTool.getLargeDataRepository();
    ProcessInstanceUUID instanceUUID = attachmentInstance.getProcessInstanceUUID();
    processInstance.addAttachment(attachmentInstance);
    Date versionDate = attachmentInstance.getVersionDate();
    String name = Misc.getAttachmentIndexName(attachmentInstance.getName(), versionDate);
    List<String> categories = Misc.getAttachmentCategories(instanceUUID);
    largeDataRepository.storeData(categories, name, value, false);
  }

  private static Execution createProcessInstance(InternalProcessDefinition process, ProcessInstanceUUID rootInstanceUUID, final ActivityDefinitionUUID activityUUID) {
    ProcessDefinitionUUID processUUID = process.getUUID();
    final long instanceNb = EnvTool.getUUIDService().getNewProcessInstanceNb(processUUID);
    final ProcessInstanceUUID instanceUUID = new ProcessInstanceUUID(processUUID, instanceNb);

    ProcessInstanceUUID theRootInstanceUUID = null;
    if (rootInstanceUUID == null) {
      theRootInstanceUUID = instanceUUID;
    } else {
      theRootInstanceUUID = rootInstanceUUID;
    }
    InternalProcessInstance instance = new InternalProcessInstance(instanceUUID, process, theRootInstanceUUID, instanceNb);
    Execution processInstance = instance.getRootExecution();

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("creating new execution for process '" + process.getName() + "'");
    }
    return processInstance;
  }

}
